import copy

from .area import Area
from ..fields import Field


class RepeatingArea:
    """
    Represents multiple identical Areas on a page.

    A root argument is required, which is expected to be a non-unique Field
    on the page.

    A collection of Areas are built from every instance of the root that is
    found. Every other Field provided in the arguments is populated inside
    each Area.

    In the following example, there's a table with 15 rows. Each row has
    two cells. The sixth row in the table should have an item with the
    name "Banana" and a price of "$7.00"

    >>> from stere.areas import RepeatingArea
    >>> from stere.fields import Root, Link, Text
    >>>
    >>> class Inventory():
    >>>     def __init__(self):
    >>>         self.inventory_items = RepeatingArea(
    >>>             root=Root('xpath', '//table/tr'),
    >>>             name=Link('xpath', './td[1]'),
    >>>             price=Text('xpath', './td[2]'),
    >>>         )

    >>> inventory = Inventory()
    >>> assert 15 == len(inventory.areas)
    >>> assert "Banana" == inventory.areas[5].name
    >>> assert "$7.00" == inventory.areas[5].price
    """
    def __init__(self, **kwargs):
        if kwargs.get('root') is None:
            raise ValueError('RepeatingArea requires a Root Field.')

        self.root = kwargs['root']

        if kwargs.get('items') is not None:
            raise ValueError('"items" is a reserved parameter.')

        self.items = {}
        for k, v in kwargs.items():
            if not isinstance(v, Field):
                raise ValueError(
                    'RepeatingArea arguments can only be Field objects.',
                )
            if k is not 'root':
                self.items[k] = v
                # Field (in plural) can be accessed directly.
                setattr(self, f'{k}s', v)

    def __len__(self):
        all_roots = self.root.find_all()
        return len(all_roots)

    @property
    def areas(self):
        """Find all instances of the root,
        then return a list of Areas: one for each root.

        Returns:
            list: Collection of every Area that was found.

        Raises:
            ValueError: If no Areas were found.

        Example:

            >>> def test_stuff():
            >>>     listings = MyPage().my_repeating_area.areas
            >>>     listings[0].my_input.fill('Hello world')

        """
        created_areas = []
        all_roots = self.root.find_all()
        if 0 == len(all_roots):
            raise ValueError(
                f'Could not find any Areas with the root: {self.root.locator}',
            )

        for item in all_roots:
            copy_items = copy.deepcopy(self.items)
            for field_name in copy_items.keys():
                copy_items[field_name]._element.parent_locator = item

            new_area = Area(**copy_items)
            created_areas.append(new_area)
        return created_areas

    def area_with(self, field_name, field_value):
        """Searches the RepeatingArea for a single Area where the Field's value
        matches the expected value and then returns the entire Area object.

        Arguments:
            field_name (str): The name of the field object.
            field_value (str): The value of the field object.

        Returns:
            Area

        Example:

            >>> class Inventory():
            >>>     def __init__(self):
            >>>         self.items = RepeatingArea(
            >>>             root=Root('xpath', '//my_xpath_string'),
            >>>             description=Text('xpath', '//my_xpath_string')
            >>>         )
            >>>
            >>> def test_stuff():
            >>>     inventory = Inventory()
            >>>     found_area = inventory.items.area_with(
            >>>         "description", "Bananas")

        """
        for area in self.areas:
            field = getattr(area, field_name)

            if field.value == field_value:
                return area

        raise ValueError(f'Could not find {field_value} in any {field_name}.')



import configparser
import os


def _parse_config():
    parser = configparser.ConfigParser()
    cwd = os.getcwd()
    file_path = f'{cwd}/stere.ini'
    parser.read(file_path)
    return parser


class BrowserEnabled:
    """Base class that stores attributes at the class level,
    shared by every object that inherits from this class.

    Attributes:
        browser (object): Pointer to what is driving the automation.
        url_navigation (str): Name of the function that opens a page.
        library (str): Name of the automation library to use. Default is
            splinter.
    """
    browser = None
    url_navigator = ''
    library = 'splinter'

    parser = _parse_config()
    if parser.has_section('stere'):
        if parser.has_option('stere', 'library'):
            library = parser.get('stere', 'library')



from .browserenabled import BrowserEnabled


class Page(BrowserEnabled):
    """Represents a single Web Page. The Page class is the base which all
    Page Objects should inherit from.

    Inheriting from Page is not required for Fields or Areas to work.

    All attribute calls are tried on the browser attribute.
    This allows classes inheriting from Page to act as a proxy to
    whichever browser/driver is being used.

    Using Splinter's browser.url method as an example, the following methods
    are analogous:

    >>> MyPage.url == MyPage.browser.url == browser.url

    The choice of which syntax to use depends on how you want to write your
    tests.
    """
    def __getattr__(self, val):
        """If an attribute doesn't exist, try getting it from the browser.
        """

        return getattr(self.browser, val)

    def __enter__(self):
        """Page Objects can be used as context managers."""
        return self

    def __exit__(self, *args):
        pass

    def navigate(self):
        """When the base Stere object has been given the `url_navigator`
        attribute, and a Page Object has a `page_url` attribute, the
        `navigate()` method can be called.

        This method will call the method defined in `url_navigator`,
        with `page_url`as the first parameter.

        In the following example, Stere is initialized with Splinter.

        >>> from splinter import Browser
        >>> from stere import Page
        >>>
        >>>
        >>> class Home(Page):
        >>>     def __init__(self):
        >>>         self.page_url = 'https://en.wikipedia.org/'
        >>>
        >>>
        >>> Stere.browser = Browser()
        >>> Stere.url_navigator = 'visit'
        >>>
        >>> home_page = Home()
        >>> home_page.navigate()
        """
        return getattr(self.browser, self.url_navigator)(self.page_url)



import copy
import time

from .strategy import strategy


class SplinterBase:
    def is_present(self, *args, **kwargs):
        """Checks if an element is present in the DOM.

        Takes the same arguments as Splinter's
        `is_element_present_by_xpath`
        """
        func = getattr(self.browser, f'is_element_present_by_{self.strategy}')
        return func(self.locator, *args, **kwargs)

    def is_not_present(self, *args, **kwargs):
        """Checks if an element is not present in the DOM.

        Takes the same arguments as Splinter's
        `is_element_not_present_by_xpath`
        """
        func = getattr(
            self.browser, f'is_element_not_present_by_{self.strategy}')
        return func(self.locator, *args, **kwargs)

    def is_visible(self, wait_time=2):
        """Checks if an element is present in the DOM and visible.

        Arguments:
            wait_time (int): The number of seconds to wait
        """
        end_time = time.time() + wait_time

        while time.time() < end_time:
            if self.find() and self.find().visible:
                return True
        return False

    def is_not_visible(self, wait_time=2):
        """Checks if an element is present in the DOM but not visible.

        Arguments:
            wait_time (int): The number of seconds to wait
        """
        end_time = time.time() + wait_time

        while time.time() < end_time:
            if not self.find() or (self.find() and not self.find().visible):
                return True
        return False

    def _find_all(self):
        """Find from page root."""
        func = getattr(self.browser, f'find_by_{self.strategy}')
        return func(self.locator)

    def _find_all_in_parent(self):
        """Find from inside a parent element."""
        func = getattr(self.parent_locator, f'find_by_{self.strategy}')
        return func(self.locator)


@strategy('css')
class FindByCss(SplinterBase):
    strategy = 'css'


@strategy('xpath')
class FindByXPath(SplinterBase):
    strategy = 'xpath'


@strategy('tag')
class FindByTag(SplinterBase):
    strategy = 'tag'


@strategy('name')
class FindByName(SplinterBase):
    strategy = 'name'


@strategy('text')
class FindByText(SplinterBase):
    strategy = 'text'


@strategy('id')
class FindById(SplinterBase):
    strategy = 'id'


@strategy('value')
class FindByValue(SplinterBase):
    strategy = 'value'


class FindByDataStarAttribute:
    """Strategy to find an element by an arbitrary data-* attribute."""
    def is_present(self, *args, **kwargs):
        return self.browser.is_element_present_by_xpath(
            f'.//*[@{self._data_star_attribute}="{self.locator}"]')

    def is_not_present(self, *args, **kwargs):
        return self.browser.is_element_not_present_by_xpath(
            f'.//*[@{self._data_star_attribute}="{self.locator}"]')

    def _find_all(self):
        """Find from page root."""
        return self.browser.find_by_xpath(
            f'.//*[@{self._data_star_attribute}="{self.locator}"]')

    def _find_all_in_parent(self):
        """Find from inside parent element."""
        return self.parent_locator.find_by_xpath(
            f'.//*[@{self._data_star_attribute}="{self.locator}"]')


def add_data_star_strategy(data_star_attribute):
    """Adds a new splinter strategy that finds by data_star_attribute.
    Args:
        data_star_attribute (str): The data-* attribute to use in the new
            strategy.
    """
    find_by_data_star = copy.deepcopy(FindByDataStarAttribute)
    find_by_data_star._data_star_attribute = data_star_attribute
    return strategy(data_star_attribute)(find_by_data_star)



from stere import Stere
from stere.strategy import strategies


def build_element(desired_strategy, locator, parent_locator=None):
    """Build a Field Object out of a BaseElement and a Strategy."""

    if desired_strategy in strategies.keys():
        bases = (BaseElement, strategies[desired_strategy])
        element_class = type('Element', bases, dict())
        return element_class(desired_strategy, locator, parent_locator)

    raise ValueError(f'The strategy "{desired_strategy}" is undefined.')


class BaseElement(Stere):
    """Combined with a Strategy to form an Element.

    Elements are used by Fields as the source of what finds the DOM element.
    """
    def __init__(self, strategy, locator, parent_locator=None):
        self.strategy = strategy
        self.locator = locator
        self.parent_locator = parent_locator

        # A Field that should be searched for and set as the parent, but only
        # when .find() is called.
        self.root = None

    def find(self):
        if self.root is None and self.parent_locator is None:
            return self._find_all()

        if self.root:
            self.parent_locator = self.root.find()
        return self._find_all_in_parent()



from .button import Button
from ..field import Field, stere_performer, use_after, use_before


@stere_performer('select', consumes_arg=True)
class Dropdown(Field):
    """Represents a dropdown menu.
    If the "option" argument is provided with a field,
        use that as the dropdown item.
    Else, assume a standard HTML Dropdown and use the option tag.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # If no option arg is given, assume Dropdown is a standard HTML one.
        if kwargs.get('option') is None:
            self.option = Button('tag', 'option')
        else:
            self.option = kwargs.get('option')

    @property
    def options(self):
        """Searches for all the elements that are an option in the dropdown.

        Returns:
            list
        """
        self.option._element.parent_locator = self.find()
        return [item for item in self.option.find_all()]

    @use_after
    @use_before
    def select(self, value):
        """Searches for an option by its html content, then clicks the one
        that matches.

        Arguments:
            value (str): The option value to select.

        Raises:
            ValueError: The provided value could not be found in the dropdown.
        """
        found_options = []
        for option in self.options:
            found_options.append(option.html)
            if option.html == value:
                option.click()
                break
        else:
            raise ValueError(
                f'{value} was not found. Found values are: {found_options}')



from functools import wraps

from .element_builder import build_element


def stere_performer(method_name, consumes_arg=False):
    """Wraps a Class that contains a method which should be
    used by Area.perform().
    """

    def wrapper(cls):
        class Performer(cls):
            def perform(self, value=None):
                performer = getattr(self, method_name)
                if consumes_arg:
                    performer(value)
                    return True
                else:
                    performer()
                    return False
        # Preserve original class name
        Performer.__name__ = cls.__name__
        return Performer
    return wrapper


def use_before(func, *args, **kwargs):
    @wraps(func)
    def wrapper(obj, *inner_args, **inner_kwargs):
        obj.before()
        return func(obj, *inner_args, **inner_kwargs)
    return wrapper


def use_after(func, *args, **kwargs):
    @wraps(func)
    def wrapper(obj, *inner_args, **inner_kwargs):
        result = func(obj, *inner_args, **inner_kwargs)
        obj.after()
        return result
    return wrapper


class Field:
    """Field objects represent individual pieces on a web page.
    Conceptually, they're modelled after general behaviours, not specific
    HTML elements.

    Arguments:
        strategy (str): The type of strategy to use when locating an element.
        locator (str): The locator for the strategy.
        workflows (list): Any workflows the Field should be included with.


    Example:

        >>> from stere.fields import Field
        >>> my_field = Field('xpath', '//*[@id="js-link-box-pt"]/small/span')

    """
    def __init__(self, strategy, locator, *args, **kwargs):
        self.strategy = strategy
        self.locator = locator
        self._element = build_element(strategy, locator)

        self.workflows = kwargs.get('workflows') or []

    def __getattr__(self, val):
        """If an attribute doesn't exist, try getting it from the element.

        If it still doesn't exist, do a find() on the element and see if the
        attribute exists there.
        """
        element = super().__getattribute__('_element')
        try:
            return getattr(element, val)
        except AttributeError:
            # Allows deepcopy not to get into infinite recursion.
            if val in ['__deepcopy__', '__getstate__']:
                raise AttributeError

            # Try getting the attribute from the found element.
            return getattr(self.find(), val)

    def __repr__(self):
        return (
            f'{self.__class__.__name__} - '
            f'Strategy: {self.strategy}, Locator: {self.locator}')

    @property
    def element(self):
        """Tries to find the element, the returns the results.
        """
        return self._element.find()

    def before(self):
        """Called automatically before methods with the `@use_before`
        decorator are called.

        By default it does nothing. Override this method if an action must be
        taken before the method has been called.
        """
        pass

    def after(self):
        """Called automatically before methods with the `@use_after`
        decorator are called.

        By default it does nothing. Override this method if an action must be
        taken after the method has been called.
        """
        pass

    @use_after
    @use_before
    def perform(self, value=None, *args, **kwargs):
        """Will be called by Area.perform()

        Returns:
            bool: True if the action used an argument, else False
        """
        return False

    def includes(self, value):
        """Will search every element found by the Field for a value property
        that matches the given value.
        If an element with a matching value is found, it's then returned.

        Useful for when you have non-unique elements and know a value is in
        one of the elements, but don't know which one.

        Arguments:
            value (str): A text string inside an element you want to find.

        Returns:
            element

        Example:

            >>> class PetStore(Page):
            >>>     def __init__(self):
            >>>         self.inventory_list = Link('xpath', '//div')
            >>>
            >>> pet_store = PetStore()
            >>> pet_store.inventory_list.includes("Kittens").click()

        """
        for item in self.element:
            if item.value == value:
                return item

    def find(self):
        """Find the first matching element.

        Returns:
            Element

        Raises:
            ValueError - If more than one element is found.
        """
        found_elements = self._element.find()
        if len(found_elements) > 2:
            raise ValueError("Expected one element, found multiple")
        return found_elements[0]

    def find_all(self):
        """Find all matching elements.

        Returns:
            List
        """
        return self._element.find()



from ..field import Field, stere_performer, use_after, use_before


@stere_performer('click', consumes_arg=False)
class Link(Field):
    """Convenience Class on top of Field.

    Uses Splinter's click method.
    """
    @use_after
    @use_before
    def click(self):
        """Uses Splinter's click method.

        Example:

            >>> login = Link('id', 'loginLink')
            >>> login.click()

        """
        self.find().click()



from ..field import Field, stere_performer, use_after, use_before


@stere_performer('fill', consumes_arg=True)
class Input(Field):
    """Convenience Class on top of Field.

    Uses Splinter's input method.
    """
    @use_after
    @use_before
    def fill(self, value=None):
        """Uses Splinter's fill method.

        Arguments:
            value (str): The text to enter into the input.

        Example:

            >>> first_name = Input('id', 'fillme')
            >>> first_name.fill('Joseph')

        """
        self.find().fill(value)
